//-*-c++-*------------------------------------------------------------
//
// File name : bioArithIntegral.cc
// Author :    Michel Bierlaire
// Date :      Tue Mar 30 17:33:15 2010
//
//--------------------------------------------------------------------

#include <sstream>

#include "patMath.h"
#include "bioParameters.h"
#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithIntegral.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithDivide.h"
#include "bioLiteralRepository.h"
#include "bioArithNotEqual.h"
#include "bioArithGaussHermite.h"
#include "bioGaussHermite.h"
#include "bioExpressionRepository.h"
#include "bioArithCompositeLiteral.h"

bioArithIntegral::bioArithIntegral(bioExpressionRepository* rep,
				   patULong par,
				   patULong left,
				   patString aLiteralName, 
				   patError*& err)
  : bioArithUnaryExpression(rep, par,left,err), literalName(aLiteralName) {
  pair<patULong,patULong> ids = 
    bioLiteralRepository::the()->getRandomVariable(literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  literalId = ids.second ;

}


bioArithIntegral::~bioArithIntegral() {}

patString bioArithIntegral::getOperatorName() const {
  return ("Integral") ;
}

patReal bioArithIntegral::getValue(patBoolean prepareGradient, patError*& err) {

  bioArithGaussHermite theGh(child,vector<patULong>(),literalId,patFALSE) ;
  bioGaussHermite theGhAlgo(&theGh) ;
  vector<patReal> result = theGhAlgo.integrate(err) ;
#ifdef BIO_CHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal() ;
  }
#endif
  return result[0] ;
}


bioExpression* bioArithIntegral::getDerivative(patULong aLiteralId, 
					       patError*& err) const {

  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    WARNING(getInfo()) ;
    return NULL;
  }
  patString derivName = bioLiteralRepository::the()->getName(aLiteralId,err) ;

  if (aLiteralId == literalId) {
    stringstream str ;
    str << "Cannot compute the derivative with respect to " << derivName << " of an integral over " << literalName ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  if (!dependsOf(aLiteralId) || child->isStructurallyZero()) {
    bioExpression* result = theRepository->getZero() ;
    return result ;
  }
  
  bioExpression* leftDeriv = child->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }



  bioExpression* result = new bioArithIntegral(theRepository,
					       patBadId,
					       leftDeriv->getId(),
					       literalName,
					       err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return result ;
}


bioArithIntegral* bioArithIntegral::getDeepCopy(bioExpressionRepository* rep,
						patError*& err) const {
  bioExpression* leftClone(NULL) ;
  if (child != NULL) {
    leftClone = child->getDeepCopy(rep,err) ;
  }
  bioArithIntegral* theNode = 
    new bioArithIntegral(rep,
			 patBadId,
			 leftClone->getId(),
			 literalName,
			 err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}


bioArithIntegral* bioArithIntegral::getShallowCopy(bioExpressionRepository* rep,
						patError*& err) const {
  bioArithIntegral* theNode = 
    new bioArithIntegral(rep,
			 patBadId,
			 child->getId(),
			 literalName,
			 err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}

patString bioArithIntegral::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for an integral") ;
  WARNING(err->describe());
  return patString() ;
}


patString bioArithIntegral::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
    if (child == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
    stringstream str ;
    str << "    bioGslParams theParams_" << prefix << getId() << " ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__x = __x ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__draw = __draw ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__beta = __beta ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".sample = sample ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__theThreadSpan = __theThreadSpan ;" << endl ;
    str << "    bioMyGhFunction_"<<prefix << getId()<<" theFunction_"<<prefix << getId()<<"(&theParams_"<<prefix << getId()<<") ;" << endl ;

    str << "    bioGaussHermite gh_"<<prefix << getId()<<"(&theFunction_"<<prefix << getId()<<") ;" << endl ;
    
    str << getCppResultDeclaration(prefix) << " = gh_"<<prefix << getId()<<".integrate() ;" << endl ;
    //    str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
    codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patBoolean bioArithIntegral::isStructurallyZero() const {
  return (child->isStructurallyZero()) ;
}

patString bioArithIntegral::getExpressionString() const {
  stringstream str ;
  str << "Integ" ;
  if (child != NULL) {
    str << '{' << child->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}

void bioArithIntegral::generateGslFunctions(ostream& str_h,patString prefix, patError*& err) {

  // Note: at this point, nested gsl functions are not
  // allowed. Therefore, the call is not propagated to the child.

  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  

  str_h << "    class bioMyGhFunction_"<< prefix << getId() <<": public bioGhFunction {" << endl ;
    str_h << "    public:" << endl ;
    str_h << "      bioMyGhFunction_"<< prefix << getId()<<"(bioGslParams* p): theParams(p) {} ;" << endl ;
    str_h << "      patReal getValue(patReal x) ;" << endl ;
    str_h << "    private:" << endl ;
    str_h << "      bioGslParams* theParams ;" << endl ;
    str_h << "    };" << endl ;
  ofstream mk("files.mk",ios::app) ;
  stringstream f_cc ;
  f_cc << "bioGslBiogeme" << prefix << getId() << ".cc" ;
  patString fileName(f_cc.str()) ;
  mk << fileName << " " ;
  mk.close() ;
  ofstream str_cc(fileName.c_str()) ;
  str_cc << "#include\"patDisplay.h\"" << endl ;
  str_cc << "#include\"bioGslParams.h\"" << endl ;
  str_cc << "#include\"bioGslBiogeme.h\"" << endl ;
  str_cc << "#include\"bioRowIterator.h\"" << endl ;
  str_cc << "#include\"bioSample.h\"" << endl ;
  //  str_cc << " #define DEBUG_MESSAGE(message)" << endl ;
  //str_cc << " #define DEBUG_MESSAGE(message) std::cout << message << endl ;" << endl ;

  patString fct = child->getGslFunction(prefix, literalId, getId() ,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  str_cc << fct << endl ;
  str_cc.close() ;
  return ;
}


bioFunctionAndDerivatives* bioArithIntegral::getNumericalFunctionAndGradient(vector<patULong> literalIds, patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  bioArithGaussHermite theGh(child,literalIds,literalId,patTRUE) ;
  bioGaussHermite theGhAlgo(&theGh) ;
  vector<patReal> r = theGhAlgo.integrate(err) ;
#ifdef BIO_CHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
#endif
  result.theFunction = r[0] ;
  for (patULong j = 0 ; j < literalIds.size() ; ++j) {
    if (finite(r[j+1])) {
      result.theGradient[j] = r[j+1] ;
    }
    else {
      result.theGradient[j] = patMaxReal ;
    }
  }
  return &result ;
}
