//-*-c++-*------------------------------------------------------------
//
// File name : bioArithLogit.cc
// Author :    Michel Bierlaire
// Date :      Sat May 21 11:23:29 2011
//
//--------------------------------------------------------------------

#include "patMath.h"
#include "bioArithLogit.h"
#include "patErrMiscError.h"
#include "patErrNullPointer.h"
#include "patDisplay.h"
#include "bioExpressionRepository.h"
#include "bioLiteralRepository.h"
#include "bioArithCompositeLiteral.h"
#include "bioArithElem.h"
#include "bioArithBinaryMinus.h"
#include "bioArithUnaryMinus.h"
#include "bioArithExp.h"
#include "bioArithMult.h"
#include "bioArithDivide.h"
#include "bioArithMultinaryPlus.h"

bioArithLogit::bioArithLogit(bioExpressionRepository* rep,
			     patULong par,
			     patULong ind,
			     map<patULong,patULong> aDict,
			     map<patULong,patULong> avDict,
			     patError*& err) :
  bioExpression(rep,par),
  indexCalculationId(ind),
  theDictionaryIds(aDict),
  theAvailDictIds(avDict) {

  
  if (theDictionaryIds.size() != theAvailDictIds.size()) {
    stringstream str ;
    str << "Incompatible dictionaries in logit. Size " 
	<< theDictionaryIds.size() << " and " << theAvailDictIds.size() ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return ;
  }

  for (map<patULong,patULong>::iterator i = theDictionaryIds.begin() ;
       i != theDictionaryIds.end() ;
       ++i) {
    // Check if both dictionaries are compatible
    map<patULong,patULong>::iterator found = theAvailDictIds.find(i->first) ;
    if (found == theAvailDictIds.end()) {
      stringstream str ;
      str << "Logit: key " << i->first << " not present in availability dictionary" ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return ;
    }
    bioExpression* theExpression = theRepository->getExpression(i->second) ;
    if (theExpression != NULL) {
      theDictionary[i->first] = theExpression ;
      theExpression->setParent(getId(),err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    else {
      stringstream str ;
      str << "Expression " << i->second << " unknown" ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return ;
    }
  }
  for (map<patULong,patULong>::iterator i = theAvailDictIds.begin() ;
       i != theAvailDictIds.end() ;
       ++i) {
    // Check if both dictionaries are compatible
    map<patULong,patULong>::iterator found = theDictionaryIds.find(i->first) ;
    if (found == theDictionaryIds.end()) {
      stringstream str ;
      str << "Logit: key " << i->first << " not present in utility dictionary" ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return ;
    }
    bioExpression* theExpression = theRepository->getExpression(i->second) ;
    if (theExpression != NULL) {
      theAvailDict[i->first] = theExpression ;
      theExpression->setParent(getId(),err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    else {
      stringstream str ;
      str << "Expression " << i->second << " unknown" ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return ;
    }
  }
  indexCalculation = theRepository->getExpression(indexCalculationId) ;
  indexCalculation->setParent(getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  }

bioArithLogit::~bioArithLogit() {

}

patString bioArithLogit::getOperatorName() const {
  return ("logit") ;
}

patReal bioArithLogit::getValue(patBoolean prepareGradient, patError*& err) {
#ifdef BIO_CHECK
  if (indexCalculation == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
#endif
  index = patULong(indexCalculation->getValue(patFALSE, err));
#ifdef BIO_CHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal() ;
  }
#endif  
  vector<patBoolean> avail ;
  for (map<patULong,bioExpression*>::const_iterator i = theAvailDict.begin() ;
       i != theAvailDict.end() ;
       ++i) {
#ifdef BIO_CHECK
    if (i->second == NULL) {
      stringstream str ;
      str << "Null availability expression in dictionary with entry " << i->first ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patReal() ;
    }
#endif
    patReal a = i->second->getValue(patFALSE,err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal() ;
    }
#endif
    avail.push_back(a != 0.0) ;
  }
  

    
  vector<patReal> Vi ;
  patULong iChosen(patBadId) ;
  patReal chosen ;
  
  patULong k = 0 ;
  for (map<patULong,patULong>::const_iterator i = theDictionaryIds.begin() ;
       i != theDictionaryIds.end() ;
       ++i) {
    bioExpression* theExpression = theRepository->getExpression(i->second) ;
    if (avail[k]) {
#ifdef BIO_CHECK
      if (theExpression == NULL) {
	stringstream str ;
	str << "Null  expression in dictionary with entry " << i->first ;
	err = new patErrMiscError(str.str()) ;
	WARNING(err->describe()) ;
	return patReal() ;
      }
#endif
      patReal V = theExpression->getValue(patFALSE,err) ;
#ifdef BIO_CHECK
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patReal() ;
	}
#endif
	if (i->first == index) {
	  chosen = V ;
	  iChosen = Vi.size() ;
	}
	Vi.push_back(V) ;
    }
    ++k ;
  }
  if (iChosen == patBadId) {
    stringstream str ;
    map<patULong,bioExpression*>::iterator found = theDictionary.find(index) ;
    if (found == theDictionary.end()) {
      stringstream str ;
      str << "Alternative " << index << " unknown in dictionary." ;
      //      WARNING(str.str()) ;
    }
    else {
      // Chosen alternative is unavailable
      WARNING("Chosen alternative is not available") ;
    }
    return 0.0 ;
  }
  
  // At this point, Vig contains the utilities (and their
  // derivatives) of the available alternatives
  
  patReal denominator(0.0) ;
  for (vector<patReal>::iterator i = Vi.begin() ;
       i != Vi.end() ;
       ++i) {
    denominator += exp(*i-chosen) ;
  }
  return(1.0 / denominator) ;
}

bioExpression* bioArithLogit::getDerivative(patULong aLiteralId, patError*& err) const {
  bioExpression* zero = theRepository->getZero() ;

  
  bioArithElem* theChosen = new bioArithElem(theRepository,
					     patBadId,
					     indexCalculation->getId(),
					     theDictionaryIds,
					     zero->getId(),err);

    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  
  
  map<patULong,bioExpression*> deriv ;
  map<patULong,patULong> derivIds ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    deriv[e->first] = e->second->getDerivative(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    derivIds[e->first] = deriv[e->first]->getId() ;
  }
  bioArithElem* theDerivChosen = new bioArithElem(theRepository,
						  patBadId,
						  indexCalculation->getId(),
						  derivIds,
						  zero->getId(),err);
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  

  vector<patULong> theSum ;
  vector<patULong> theSumDeriv ;

  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    bioArithBinaryMinus* theDiff = 
      new bioArithBinaryMinus(theRepository,
			      patBadId,
			      e->second->getId(),
			      theChosen->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    bioArithExp* theExp = 
      new bioArithExp(theRepository,patBadId,theDiff->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    map<patULong,patULong>::const_iterator found = theAvailDictIds.find(e->first) ;
    if (found == theAvailDictIds.end()) {
      err = new patErrMiscError("Incompatible dictionaries") ;
      WARNING(err->describe()) ;
      return NULL ;
    }

 
    bioArithMult* theAvailExp = 
      new bioArithMult(theRepository,patBadId,theExp->getId(),found->second,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    theSum.push_back(theAvailExp->getId()) ;

    bioExpression* theDerivDiff ;
    if (theDerivChosen->isStructurallyZero()) {
      if (deriv[e->first]->isStructurallyZero()) {
	theDerivDiff = zero ;
      }
      else {
	theDerivDiff = deriv[e->first];
      }
    }
    else {
      if (deriv[e->first]->isStructurallyZero()) {
	theDerivDiff = new bioArithUnaryMinus(theRepository,patBadId,theDerivChosen->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
      }
      else {
	theDerivDiff = new bioArithBinaryMinus(theRepository,
					       patBadId,
					       deriv[e->first]->getId(),
					       theDerivChosen->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
      }

    }

    

    bioExpression* theDerivProd ;
    if (theAvailExp->isStructurallyZero() || theDerivDiff->isStructurallyZero()) {
      theDerivProd = zero ;
    }
    else {
      theDerivProd = new bioArithMult(theRepository,patBadId,theAvailExp->getId(),theDerivDiff->getId(),err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
    }
    theSumDeriv.push_back(theDerivProd->getId()) ;
      
  }  
  
  bioArithMultinaryPlus* theDenom = new bioArithMultinaryPlus(theRepository,patBadId,theSum,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }


  bioArithMultinaryPlus* theDenomDeriv = new bioArithMultinaryPlus(theRepository,patBadId,theSumDeriv,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }


  bioArithMult* theProd = new bioArithMult(theRepository,
					   patBadId,
					   theDenom->getId(),
					   theDenom->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  bioArithDivide* theRatio = new bioArithDivide(theRepository,
						patBadId,
						theDenomDeriv->getId(),
						theProd->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  bioArithUnaryMinus* theResult = new bioArithUnaryMinus(theRepository,patBadId,theRatio->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  return theResult ;
}

bioArithLogit* bioArithLogit::getDeepCopy(bioExpressionRepository* rep,
					  patError*& err) const {
  bioExpression* indexClone(NULL) ;
  if (indexCalculation != NULL) {
    indexClone = indexCalculation->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  map<patULong,patULong> newDictionary ;
  
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    bioExpression* clone = e->second->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    newDictionary[e->first] = clone->getId() ;
  }

  map<patULong,patULong> newAv ;
  
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    bioExpression* clone = a->second->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    newAv[a->first] = clone->getId() ;
  }

  bioArithLogit* theNode = 
    new bioArithLogit(rep,patBadId,indexClone->getId(),newDictionary,newAv,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;

}


bioArithLogit* bioArithLogit::getShallowCopy(bioExpressionRepository* rep,
					  patError*& err) const {
  bioArithLogit* theNode = 
    new bioArithLogit(rep,patBadId,indexCalculationId,theDictionaryIds,theAvailDictIds,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;

}
patString bioArithLogit::getCppCode(patString prefix, patError*& err) {
  err = new patErrMiscError("Not yet implemented") ;
  WARNING(err->describe()) ;
  return patString() ;
}
patString bioArithLogit::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("Not yet implemented") ;
  WARNING(err->describe()) ;
  return patString();
}
patString bioArithLogit::getExpression(patError*& err) const {
  stringstream str ;
  str << "Logit" ;
  patString theIndex = indexCalculation->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  str << "[" << index << "]" ;
  str << "{" ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end();
       ++i) {
    if (i != theDictionary.begin()) {
      str << "," ;
    }
    patString theExpr = i->second->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    map<patULong,bioExpression*>::const_iterator found =
      theAvailDict.find(i->first) ;
    patString theAv = found->second->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    str << i->first << ": " << theExpr << "(Av: " << theAv << ")"  ;
  }
  str << "}" ;
  return patString(str.str()) ;


}
patBoolean bioArithLogit::isSimple() const {
  return patFALSE ;
}

patBoolean bioArithLogit::dependsOf(patULong aLiteralId) const {
  if (indexCalculation->dependsOf(aLiteralId)) {
    return patTRUE ;
  }
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    if (e->second->dependsOf(aLiteralId)) {
      return patTRUE ;
    }
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    if (a->second->dependsOf(aLiteralId)) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}
void bioArithLogit::forceCodeGeneration() {
  codeMustBeGenerated = patTRUE ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->forceCodeGeneration() ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->forceCodeGeneration() ;
  }
}

void bioArithLogit::simplifyZeros(patError*& err) {
  err = new patErrMiscError("Not yet implemented") ;
  WARNING(err->describe()) ;

}

patULong bioArithLogit::getNumberOfOperations() const {
  return 0 ;
}

patULong bioArithLogit::lastIndex() const {
  return theDictionary.rbegin()->first ;
}

patString bioArithLogit::getExpressionString() const {
  return patString() ;
}

patBoolean bioArithLogit::containsAnIntegral() const {
  if (indexCalculation->containsAnIntegral()) {
    return patTRUE ;
  }
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    if (e->second->containsAnIntegral()) {
      return patTRUE ;
    }
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    if (a->second->containsAnIntegral()) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}
patBoolean bioArithLogit::containsAnIterator() const {
  if (indexCalculation->containsAnIterator()) {
    return patTRUE ;
  }
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    if (e->second->containsAnIterator()) {
      return patTRUE ;
    }
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    if (a->second->containsAnIterator()) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}
patBoolean bioArithLogit::containsASequence() const {
  if (indexCalculation->containsASequence()) {
    return patTRUE ;
  }
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    if (e->second->containsASequence()) {
      return patTRUE ;
    }
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    if (a->second->containsASequence()) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}

void bioArithLogit::generateGslFunctions(ostream& str_h,patString prefix,patError*& err) {
  err = new patErrMiscError("Not yet implemented") ;
  WARNING(err->describe()) ;

}
patString bioArithLogit::getIncludeStatements(patString prefix) const {
  return patString("Not yet implemented") ;
}
void bioArithLogit::setVariables(const patVariables* x) {
  __x = x ;
  indexCalculation->setVariables(x) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->setVariables(x) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->setVariables(x) ;
  }
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setVariables(x) ;
  }
}

void bioArithLogit::setDraws(const vector< vector<patReal> >* d) {
  __draws = d ;
  indexCalculation->setDraws(d) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->setDraws(d) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->setDraws(d) ;
  }
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setDraws(d) ;
  }
}

void bioArithLogit::setSample(const bioSample* s) {
  theSample = s ;
  indexCalculation->setSample(s) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->setSample(s) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->setSample(s) ;
  }
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setSample(s) ;
  }
}

void bioArithLogit::setCurrentSpan(bioIteratorSpan aSpan) {
  theCurrentSpan = aSpan ;
  indexCalculation->setCurrentSpan(aSpan) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->setCurrentSpan(aSpan) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->setCurrentSpan(aSpan) ;
  }
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setCurrentSpan(aSpan) ;
  }
}

void bioArithLogit::setThreadSpan(bioIteratorSpan aSpan) {
  theThreadSpan = aSpan ;
  indexCalculation->setThreadSpan(aSpan) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->setThreadSpan(aSpan) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->setThreadSpan(aSpan) ;
  }
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setThreadSpan(aSpan) ;
  }
}

void bioArithLogit::collectExpressionIds(set<patULong>* s) const {
  s->insert(getId()) ;
  indexCalculation->collectExpressionIds(s) ;
  for (map<patULong,bioExpression*>::const_iterator e = theDictionary.begin() ;
       e != theDictionary.end() ;
       ++e) {
    e->second->collectExpressionIds(s) ;
  }
  for (map<patULong,bioExpression*>::const_iterator a = theAvailDict.begin() ;
       a != theAvailDict.end() ;
       ++a) {
    a->second->collectExpressionIds(s) ;
  }
  for (vector<bioExpression*>::const_iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->collectExpressionIds(s) ;
  }
}



bioFunctionAndDerivatives* bioArithLogit::getNumericalFunctionAndGradient(vector<patULong> literalIds, patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }

#ifdef BIO_CHECK
  if (indexCalculation == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
#endif
  index = patULong(indexCalculation->getValue(patFALSE, err));
#ifdef BIO_CHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
#endif  

  vector<patBoolean> avail ;
  for (map<patULong,bioExpression*>::const_iterator i = theAvailDict.begin() ;
       i != theAvailDict.end() ;
       ++i) {
    patReal a = i->second->getValue(patFALSE,err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
#endif
    avail.push_back(a != 0.0) ;
  }
    
  Vig.erase(Vig.begin(),Vig.end()) ;
  patULong iChosen(patBadId) ;
  bioFunctionAndDerivatives chosen(literalIds.size()) ;
    
  patULong k = 0 ;
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    if (avail[k]) {
      bioFunctionAndDerivatives* V = i->second->getNumericalFunctionAndGradient(literalIds,err) ;
#ifdef BIO_CHECK
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
#endif
      if (i->first == index) {
	chosen = *V ;
	iChosen = Vig.size() ;
      }
      Vig.push_back(V) ;
    }
    ++k ;
  }
  if (iChosen == patBadId) {
    stringstream str ;
    map<patULong,bioExpression*>::iterator found = theDictionary.find(index) ;
    if (found == theDictionary.end()) {
      // str << "Alternative " << index << " unknown in dictionary." ;
      // WARNING(err->describe()) ;
    }
    else {
      // str << "Cannot compute logit for alternative " 
      // 	  << index 
      // 	  << " as it is declared unavailable" ;
      // err = new patErrMiscError(str.str()) ;
      // WARNING(err->describe()) ;
    }
    result.theFunction = 0.0 ;
    for (patULong j = 0 ; j < literalIds.size() ; ++j) {
	result.theGradient[j] = 0.0 ;
    }
    return &result ;
  }
  
  // At this point, Vig contains the utilities (and their
  // derivatives) of the available alternatives

  vector<patReal> expi ;
    
  vector<patReal> dprime(literalIds.size(),0.0) ;
  patReal denominator(0.0) ;
  for (vector<bioFunctionAndDerivatives*>::iterator i = Vig.begin() ;
       i != Vig.end() ;
       ++i) {
    patReal e = exp((*i)->theFunction-chosen.theFunction) ;
    for (patULong j = 0 ; j < literalIds.size() ; ++j) {
      if ((*i)->theGradient[j] != chosen.theGradient[j]) {
	dprime[j] += e * ((*i)->theGradient[j] - chosen.theGradient[j]) ;
      }
    }
    denominator += e ;
  }
  patReal pi ;
  if (patAbs(denominator) < patEPSILON) {
    pi = patMaxReal ;
  }
  else {
    pi = 1.0 / denominator ;
  }
  result.theFunction = pi ;
  if (pi == patMaxReal) {
    for (patULong j = 0 ; j < literalIds.size() ; ++j) {
	result.theGradient[j] = patMaxReal ;
    }
  }
  else {
    for (patULong j = 0 ; j < literalIds.size() ; ++j) {
      if (dprime[j] == 0) {
	result.theGradient[j] = 0.0 ;
      }
      else {
	result.theGradient[j] = - dprime[j] * pi * pi ;
      }
    }    
  }
#ifdef BIO_CHECK
  for (patULong i = 0 ; i < literalIds.size() ; ++i) {
    if (!finite(result.theGradient[i])) {
      result.theGradient[i] = patMaxReal ;
    }
  }
#endif
  return &result ;
}

patString bioArithLogit::check(patError*& err) const  {
  stringstream str ;
  if (indexCalculation->getId() != indexCalculationId) {
      str << "Incompatible IDS for children: " << indexCalculation->getId() << " and " << indexCalculationId << endl ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  if (indexCalculation->getParent() != getId()) {
    str << "Incompatible IDS for children: " << indexCalculation->getParent() << " and " << getId() << endl ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;

  }
  for (map<patULong,bioExpression*>::const_iterator i = theDictionary.begin() ;
       i != theDictionary.end() ;
       ++i) {
    map<patULong,patULong>::const_iterator found = theDictionaryIds.find(i->first) ;
    if (found == theDictionaryIds.end()) {
      str << "Incompatible data structure for bioArithElem: expression " << i->first << " does not have an associated ID" << endl ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;
    }
    if (i->second->getId() != found->second) {
      str << "Incompatible IDS for children: " << i->second->getId() << " and " << found->second << endl ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    if (i->second->getParent() != getId()) {
      str << "Incompatible IDS for children: " << i->second->getParent() << " and " << getId() << endl  ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString() ;
    }
  }

  for (map<patULong,bioExpression*>::const_iterator i = theAvailDict.begin() ;
       i != theAvailDict.end() ;
       ++i) {
    map<patULong,patULong>::const_iterator found = theAvailDictIds.find(i->first) ;
    if (found == theAvailDictIds.end()) {
      str << "Incompatible data structure for bioArithElem: expression " << i->first << " does not have an associated ID" << endl ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    if (i->second->getId() != found->second) {
      str << "Incompatible IDS for children: " << i->second->getId() << " and " << found->second << endl ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    if (i->second->getParent() != getId()) {
      str << "Incompatible IDS for children: " << i->second->getParent() << " and " << getId() << endl  ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString() ;
    }
  }

  return patString(str.str()) ;
}

