//-*-c++-*------------------------------------------------------------
//
// File name : bioArithProd.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Wed Jul  8 15:07:35  2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "patError.h"

#include "bioArithProd.h"
#include "bioArithMult.h"
#include "bioArithDivide.h"
#include "bioArithSum.h"
#include "bioLiteralRepository.h"
#include "bioArithCompositeLiteral.h"
#include "bioRandomDraws.h"
#include "bioSample.h"
#include "bioRowIterator.h"
#include "bioMetaIterator.h"
#include "bioDrawIterator.h"

bioArithProd::bioArithProd(bioExpressionRepository* rep,
			   patULong par,
			   patULong left,
			   patString it,
			   patError*& err) 
  : bioArithIterator(rep, par,left,it,err)  {
  
  if (err != NULL) {
    WARNING(err->describe()) ;
  }
}


bioArithProd::~bioArithProd() {}


patString bioArithProd::getOperatorName() const {
  return patString("Prod") ;
}


bioExpression* bioArithProd::getDerivative(patULong aLiteralId, patError*& err) const {

  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* theProd = new bioArithProd(theRepository,patBadId,child->getId(),theIteratorName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* leftValue = child->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* leftResult = child->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* ratio = new bioArithDivide(theRepository,patBadId,leftResult->getId(),leftValue->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    bioExpression* theSum = new bioArithSum(theRepository,patBadId,ratio->getId(),theIteratorName,patBadId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* result = new bioArithMult(theRepository,patBadId,theProd->getId(),theSum->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  return result ;
}


bioArithProd* bioArithProd::getDeepCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioExpression* leftClone(NULL) ;
  if (child != NULL) {
    leftClone = child->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  bioArithProd* theNode = 
    new bioArithProd(rep,patBadId,leftClone->getId(),theIteratorName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  return theNode ;
}

bioArithProd* bioArithProd::getShallowCopy(bioExpressionRepository* rep,
					patError*& err) const {
  bioArithProd* theNode = 
    new bioArithProd(rep,patBadId,child->getId(),theIteratorName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  return theNode ;
}

patBoolean bioArithProd::isSum() const {
  return patFALSE ;
}

patBoolean bioArithProd::isProd() const {
  return patTRUE ;
}

patString bioArithProd::getExpressionString() const {
  stringstream str ;
  str << "$P" ;
  str << theIteratorName ;
  if (child != NULL) {
    str << '{' << child->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;
}


patReal bioArithProd::getValue(patBoolean prepareGradient, 
			       patError*& err) {
#ifdef BIO_CHECK
  if (child->getId() != childId) {
    stringstream str ;
    str << "Ids are diffentent: " 
	<< child->getId() 
	<< " <> " 
	<< childId ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
  
  
  if (theSample == NULL) {
    err = new patErrNullPointer("bioSample") ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
#endif
  patReal result = 1.0 ;
  
  if (theIteratorType == ROW) {
    bioRowIterator* theIter = theSample->createRowIterator(theCurrentSpan,
							   theThreadSpan,
							   err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal() ;
    }
#endif
    patULong nIter(0) ;
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      ++nIter ;
      child->setVariables(theIter->currentItem()) ;
      patReal expression = child->getValue(patFALSE,err) ;
#ifdef BIO_CHECK
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patReal() ;
      }
#endif
      result *= expression ;
    }
    DELETE_PTR(theIter) ;
    return result ;
  }
  else if (theIteratorType == META) {
    bioMetaIterator* theIter = theSample->createMetaIterator(theCurrentSpan,theThreadSpan,err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal() ;
    }
#endif
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      child->setCurrentSpan(theIter->currentItem()) ;
      patReal expression = child->getValue(patFALSE,err) ;
#ifdef BIO_CHECK
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patReal() ;
      }
#endif
      result *= expression ;
    }
    DELETE_PTR(theIter) ;
    return result ;
  }
  else if (theIteratorType == DRAW) {
    err = new patErrMiscError("No product with draws") ;
    WARNING(err->describe()) ;
    return patReal() ;
  }
  err = new patErrMiscError("Should never be reached") ;
  WARNING(err->describe()) ;
  return patReal() ;
  return patReal() ;
}


patULong bioArithProd::getNumberOfOperations() const {
  patError* err(NULL) ;
    
  patULong result = 0 ;
    
    
  if (theIteratorType == ROW) {
    bioRowIterator* theIter = theSample->createRowIterator(theCurrentSpan,theThreadSpan,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal() ;
    }
    patULong nIter(0) ;
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      ++nIter ;
      result += child->getNumberOfOperations() ;
    }
    DELETE_PTR(theIter) ;
    return result ;
  }
  else if (theIteratorType == META) {
    bioMetaIterator* theIter = theSample->createMetaIterator(theCurrentSpan,theThreadSpan,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal() ;
    }
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      child->setCurrentSpan(theIter->currentItem()) ;
      result += child->getNumberOfOperations() ;
    }
    DELETE_PTR(theIter) ;
    return result ;
  }
  else if (theIteratorType == DRAW) {
    bioDrawIterator* theIter = bioRandomDraws::the()->createIterator() ;
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      result += child->getNumberOfOperations() ;
    }
    DELETE_PTR(theIter) ;
    // There is one more division
    return 1+result ;
  }
  err = new patErrMiscError("Should never be reached") ;
  WARNING(err->describe()) ;
  return patReal() ;
}


bioFunctionAndDerivatives* bioArithProd::getNumericalFunctionAndGradient(vector<patULong> literalIds, patError*& err) {

  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }

  patReal r = 1.0 ;
  
  vector<patReal> gradient(literalIds.size(),0.0) ;
  
  if (theIteratorType == ROW) {
    bioRowIterator* theIter = theSample->createRowIterator(theCurrentSpan,
							   theThreadSpan,
							   err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
#endif
    patULong nIter(0) ;
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      ++nIter ;
      child->setVariables(theIter->currentItem()) ;
      bioFunctionAndDerivatives* fg = child->getNumericalFunctionAndGradient(literalIds,err) ;
#ifdef BIO_CHECK
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
#endif
      r *= fg->theFunction ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	gradient[i] += fg->theGradient[i]/fg->theFunction ;
      }
    }
    DELETE_PTR(theIter) ;
    result.theFunction = r ;
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      result.theGradient[i] = r * gradient[i] ;
    }    
    return &result ;
  }
  else if (theIteratorType == META) {
    bioMetaIterator* theIter = theSample->createMetaIterator(theCurrentSpan,theThreadSpan,err) ;
#ifdef BIO_CHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
#endif
    for (theIter->first() ;
	 !theIter->isDone() ;
	 theIter->next()) {
      child->setCurrentSpan(theIter->currentItem()) ;
      bioFunctionAndDerivatives* fg = child->getNumericalFunctionAndGradient(literalIds,err) ;
#ifdef BIO_CHECK
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
#endif
      r *= fg->theFunction ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	gradient[i] += fg->theGradient[i] / fg->theFunction ;
      }
    }
    DELETE_PTR(theIter) ;
    result.theFunction = r ;
    for (patULong i = 0 ; i < literalIds.size() ; ++i) {
      result.theGradient[i] = r * gradient[i] ;
    }    
    return &result ;
  }
  else if (theIteratorType == DRAW) {
    err = new patErrMiscError("No product with draws") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  err = new patErrMiscError("Should never be reached") ;
  WARNING(err->describe()) ;
  return NULL ;
}
