"""
File name : bio_expression.py
Author :    Michel Bierlaire and Mamy Fetiarison
Date :      Fri May 15 15:06:23 2009
"""
class Operator:
  
   # Name of the operators and the functions
   # Each name must be unique
   num = 'numeric'
   var = 'variable'
   rv = 'randomVariable'
   param = 'beta'
   normal = 'N(0,1)'
   uniform = 'U[0,1]'
   uniformSym = 'U[-1,1]'
   absOp, negOp = 'abs','minus'
   exp, log, sqrt = 'exp','log','sqrt'
   bioNormalCdf = 'bioNormalCdf'
   add, sub, mul, div, power = '+', '-', '*', '/', '**'
   andOp, orOp, equal, notEqual = 'and','or','==','<>'
   greater, greaterEq, less, lessEq = '>','>=','<','<='
   minOp, maxOp, mod = 'min','max','mod'
   sumOp, prodOp, integralOp, derivativeOp = 'sum', 'prod','integral','derivative'
   elemOp, enumOp, logitOp, bioNormalPdf = 'elem', 'enumerate','logit','bioNormalPdf'
   defineOp = 'define'

   # Bounds
   MIN_ZEROOP_INDEX     =  0
   MAX_ZEROOP_INDEX     = 19
   MIN_UNOP_INDEX       = 20
   MAX_UNOP_INDEX       = 39
   MIN_BINOP_INDEX      = 40    
   MAX_BINOP_INDEX      = 69
   MIN_ITERATOROP_INDEX = 70
   MAX_ITERATOROP_INDEX = 89

   UNDEF_OP = -1

   # Each operator is associated with an index depending on
   # the above bounds
   operatorIndexDic = {num:0, \
                    var:1, \
                    param:2, \
                    normal:3, \
                    uniform:4, \
                    rv:5, \
                    uniformSym:6, \
                    absOp:20, negOp:21, \
                    exp:30, log:31, sqrt:32, bioNormalCdf: 33, \
                    add:40, sub:41, mul:42, div:43, power:44, \
                    andOp:45, orOp:46, equal:47, notEqual:48, \
                    greater:49, greaterEq:50, less:51, lessEq:52, \
                    minOp:53, maxOp:54, mod:55, \
                    sumOp:70, prodOp:71,  \
                    elemOp:90, enumOp:91,  integralOp:92, derivativeOp:93, \
                    defineOp:94, logitOp:95, bioNormalPdf:96
                   }
	
   # Return the index associated to an operator
   def getOpIndex(self,op):
      return Operator.operatorIndexDic[op];


Operator = Operator()


# Build an "Expression" object from a numeric value
def buildExpressionObj(exp):
   def isNumeric(obj):
      # Consider only ints and floats numeric
      return isinstance(obj,int) or isinstance(obj,float)

   if isNumeric(exp) :
      return Numeric(exp)
   else :
      return exp


class Expression:
   """ Interface for any expression """
   
   def __init__(self):
      self.operatorIndex = UNDEF_OP

   
   # Return the string representation of the current expression
   # Must be implemented in the sub classes
   def getExpression(self):
      raise NotImplementedError("getExpression must be implemented! ")

   def __str__(self):
      return self.getExpression()

   def __neg__(self):
      return UnOp(Operator.negOp, self)
      
   def __add__(self, expression):
      return BinOp(Operator.add, self, buildExpressionObj(expression))

   def __radd__(self, expression):
      return BinOp(Operator.add, buildExpressionObj(expression), self)

   def __sub__(self, expression):
      return BinOp(Operator.sub, self, buildExpressionObj(expression))

   def __rsub__(self, expression):
      return BinOp(Operator.sub, buildExpressionObj(expression), self)

   def __mul__(self, expression):
      return BinOp(Operator.mul, self, buildExpressionObj(expression))
      
   def __rmul__(self, expression):
      return BinOp(Operator.mul, buildExpressionObj(expression), self)

   def __div__(self, expression):
      return BinOp(Operator.div, self, buildExpressionObj(expression))

   def __rdiv__(self, expression):
      return BinOp(Operator.div, buildExpressionObj(expression), self)

   # Support for Python version 3.x
   def __truediv__(self, expression):
      return BinOp(Operator.div, self, buildExpressionObj(expression))
#      return self.__div__(expression)

   def __rtruediv__(self, expression):
      return BinOp(Operator.div, buildExpressionObj(expression), self)
#      return self.__div__(expression)
   ##########################################


   def __mod__(self, expression):
      return BinOp(Operator.modOp, self, buildExpressionObj(expression))

   def __pow__(self, expression):
      return BinOp(Operator.power, self, buildExpressionObj(expression))

   def __rpow__(self, expression):
      return BinOp(Operator.power, buildExpressionObj(expression), self)

   def __and__(self, expression):
      return BinOp(Operator.andOp, self, buildExpressionObj(expression))

   def __or__(self, expression):
      return BinOp(Operator.orOp, self, buildExpressionObj(expression))

   def __eq__(self, expression):
      return BinOp(Operator.equal, self, buildExpressionObj(expression))
	  
   def __ne__(self, expression):
      return BinOp(Operator.notEqual, self, buildExpressionObj(expression))

   def __le__(self, expression):
      return BinOp(Operator.lessEq, self, buildExpressionObj(expression))

   def __ge__(self, expression):
      return BinOp(Operator.greaterEq, self, buildExpressionObj(expression))

   def __lt__(self, expression):
      return BinOp(Operator.less, self, buildExpressionObj(expression))

   def __gt__(self, expression):
      return BinOp(Operator.greater, self, buildExpressionObj(expression))


class Numeric(Expression):
   """ Class wrapping an integer or a float value """

   def __init__(self,number):
      self.number = number
      self.operatorIndex = Operator.operatorIndexDic[Operator.num]

   def getExpression(self):
      return "(" + str(self.number) + ")"


class Variable(Expression):

   def __init__(self,name):
      self.name = name
      self.operatorIndex = Operator.operatorIndexDic[Operator.var]

   def getExpression(self):
      return str(self.name)

class RandomVariable(Expression):

   def __init__(self,name):
      self.name = name
      self.operatorIndex = Operator.operatorIndexDic[Operator.rv]

   def getExpression(self):
      return str(self.name)


class Beta(Expression):

   def __init__(self,name,val,lb,ub,st,desc=''):
      self.name = name
      self.val = val
      self.lb = lb
      self.ub = ub
      self.st = st
      self.desc = desc
      self.operatorIndex = Operator.operatorIndexDic[Operator.param]

   def getExpression(self):
      return str(self.name)
      #return self.name + " " + str(self.val) + " " + str(self.lb) + \
      #       " " + str(self.ub) + " " + str(self.st) + " " + self.desc
   

class bioNormalDraws(Expression):

   def __init__(self,name,index='__rowId__'):
      self.name = name
      # Index of the column of the sample file for which we need random numbers
      self.index = index
      self.operatorIndex = Operator.operatorIndexDic[Operator.normal]

   def getExpression(self):
      return str(self.name)


# Uniform [-1:1]
class bioUniformSymmetricDraws(Expression):

   def __init__(self,name,index='__rowId__'):
      self.name = name
      # Index of the column of the sample file for which we need random numbers
      self.index = index
      self.operatorIndex = Operator.operatorIndexDic[Operator.uniformSym]

   def getExpression(self):
      return str(self.name)

# Uniform [0:1]
class bioUniformDraws(Expression):

   def __init__(self,name,index='__rowId__'):
      self.name = name
      # Index of the column of the sample file for which we need random numbers
      self.index = index
      self.operatorIndex = Operator.operatorIndexDic[Operator.uniform]

   def getExpression(self):
      return str(self.name)



class UnOp(Expression):
   """ Unary operators """

   def __init__(self,op,expression):
      self.op = op
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[op]

   def getExpression(self):
      return self.op + "(" + self.expression.getExpression() + ")"


class abs(Expression):
   """ Absolute value """
   
   def __init__(self,expression):
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[Operator.absOp]

   def getExpression(self):
      return Operator.absOp + "(" + self.expression.getExpression() + ")"


class exp(Expression):
   """ Exponential function """
   
   def __init__(self,expression):
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[Operator.exp]

   def getExpression(self):
      return Operator.exp + "(" + self.expression.getExpression() + ")"


class log(Expression):
   """ Logarithm function """
   
   def __init__(self,expression):
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[Operator.log]

   def getExpression(self):
      return Operator.log + "(" + self.expression.getExpression() + ")"


class sqrt(Expression):
   """ Sqrt function """
   
   def __init__(self,expression):
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[Operator.sqrt]

   def getExpression(self):
      return Operator.sqrt + "(" + self.expression.getExpression() + ")"


class bioNormalCdf(Expression):
   """ normal Cdf function """
   
   def __init__(self,expression):
      self.expression = buildExpressionObj(expression)
      self.operatorIndex = Operator.operatorIndexDic[Operator.bioNormalCdf]

   def getExpression(self):
      return Operator.normalCdf + "(" + self.expression.getExpression() + ")"


class max(Expression):
   def __init__(self,left,right):
      self.left = buildExpressionObj(left)
      self.right = buildExpressionObj(right)
      self.operatorIndex = Operator.operatorIndexDic[Operator.maxOp]

   def getExpression(self):
      return "max(" + self.left.getExpression() + "," + self.right.getExpression() + ")"

class min(Expression):
   def __init__(self,left,right):
      self.left = buildExpressionObj(left)
      self.right = buildExpressionObj(right)
      self.operatorIndex = Operator.operatorIndexDic[Operator.minOp]

   def getExpression(self):
      return "min(" + self.left.getExpression() + "," + self.right.getExpression() + ")"

class BinOp(Expression):

   def __init__(self,op,left,right):
      self.op = op
      self.left = buildExpressionObj(left)
      self.right = buildExpressionObj(right)
      self.operatorIndex = Operator.operatorIndexDic[op]

   def getExpression(self):
      return "(" + self.left.getExpression() + self.op + self.right.getExpression() + ")"

class Sum(Expression) :

   def __init__(self, term, iteratorName) :
      self.function = buildExpressionObj(term)
      self.iteratorName = iteratorName 
      self.operatorIndex = Operator.operatorIndexDic[Operator.sumOp]

   def getExpression(self) :
      strexpr = "sum"
      strexpr += "[" + str(self.iteratorName) + "]"
      strexpr += "(" + self.function.getExpression() + ")"
      return strexpr


class Prod(Expression) :

   def __init__(self, term, iteratorName) :
      self.function = buildExpressionObj(term)
      self.iteratorName = iteratorName 
      self.operatorIndex = Operator.operatorIndexDic[Operator.prodOp]

   def getExpression(self) :
      strexpr = "prod"
      strexpr += "[" + str(self.iteratorName) + "]"
      strexpr += "(" + self.function.getExpression() + ")"
      return strexpr

class Integrate(Expression):
   def __init__(self, term,v):
      self.function = buildExpressionObj(term)
      self.variable = v
      self.operatorIndex = Operator.operatorIndexDic[Operator.integralOp]

   def getExpression(self) :
      strexpr = "Integral"
      strexpr += "(" + self.function.getExpression() + "," + variable + ")"
      return strexpr
                
class Derive(Expression):
   def __init__(self, term,v):
      self.function = buildExpressionObj(term)
      self.variable = v
      self.operatorIndex = Operator.operatorIndexDic[Operator.derivativeOp]

   def getExpression(self) :
      strexpr = "Derive"
      strexpr += "(" + self.function.getExpression() + "," + variable + ")"
      return strexpr
                

class bioNormalPdf(Expression):
   def __init__(self, term):
      self.function = buildExpressionObj(term)
      self.operatorIndex = Operator.operatorIndexDic[Operator.bioNormalPdf]

   def getExpression(self) :
      strexpr = "normalPdf"
      strexpr += "(" + self.function.getExpression() + ")"
      return strexpr
                

class Elem(Expression) :
   
   def __init__(self, prob, choice, default = Numeric(0)) :
      self.prob = {} # dictionary
      for k,v in prob.items() :
         self.prob[k] = buildExpressionObj(v)
         
      self.choice = buildExpressionObj(choice)
      self.default = buildExpressionObj(default)
      self.operatorIndex = Operator.operatorIndexDic[Operator.elemOp]

   def getExpression(self) :
      res = "Elem"
      res += "[" + str(self.choice) + "]"
      res += "{"
      for i,v in self.prob.items():
         res += "(" + str(i) + ": " + str(v) + ")"
      res += "}"
      return res


class bioLogit(Expression) :
   
   def __init__(self, prob, av, choice) :
      self.prob = {} # dictionary
      for k,v in prob.items() :
         self.prob[k] = buildExpressionObj(v)
      self.av = {}   
      for k,v in av.items() :
         self.av[k] = buildExpressionObj(v)
      self.choice = buildExpressionObj(choice)
      self.operatorIndex = Operator.operatorIndexDic[Operator.logitOp]

   def getExpression(self) :
      res = "Logit"
      res += "[" + str(self.choice) + "]"
      res += "{"
      for i,v in self.prob.items():
         res += "(" + str(i) + ": " + str(v) + ")"
      res += "}"
      res += "{"
      for i,v in self.av.items():
         res += "(" + str(i) + ": " + str(v) + ")"
      res += "}"
      return res


class Enumerate(Expression) :

   def __init__(self, term, iteratorName) :
      self.theDict = {}
      for k,v in term.items():
         self.theDict[k] = buildExpressionObj(v)
      self.iteratorName = iteratorName 
      self.operatorIndex = Operator.operatorIndexDic[Operator.enumOp]

   def getExpression(self) :
      strexpr = "Enumerate"
      strexpr += "[" + str(self.iteratorName) + "]"
      strexpr += "{"
      for i,v in self.term.items():
         strexpr += "(" + str(i) + ": " + str(v) + ")"
      strexpr += "}"
      return strexpr

class Define(Expression) :
   def __init__(self, name, expr, classname=None):
      self.name = name
      self.expr = expr
      self.operatorIndex = Operator.operatorIndexDic[Operator.defineOp]

   def getExpression(self):
      return str(self.name)
      
   def assign(self, expr) :
      self.expr = expr

